#include <Converter.hpp>

#include <string>

#include <QDomComment>

#include <abstraction/ImmediateValueAbstraction.hpp>
#include <abstraction/ValueProvider.hpp>
#include <common/AlgorithmHelper.h>
#include <registry/StringWriterRegistry.hpp>
#include <registry/XmlRegistry.h>
#include <sax/SaxComposeInterface.h>
#include <sax/SaxParseInterface.h>

#include <GraphvizIntegrator.hpp>

namespace Converter {

    std::optional<QString> toString(const std::shared_ptr<abstraction::OperationAbstraction>& data) {
        try {
            auto res = abstraction::StringWriterRegistry::getAbstraction(data->getReturnType());
            res->attachInput(data, 0, true, true);
            res->eval();

            auto& value = dynamic_cast<abstraction::ValueProvider<std::string>&>(*res);
            return QString::fromStdString(value.getValue(true));
        }
        catch (const exception::CommonException& e) {
            return std::nullopt;
        }
        catch (const std::exception& e) {
            return std::nullopt;
        }
    }

    std::optional<QString> toDOT(const std::shared_ptr<abstraction::OperationAbstraction>& data) {
        try {
            ext::vector<std::shared_ptr<abstraction::OperationAbstraction>> params;
            params.push_back(data);

            std::vector<bool> moves;
            moves.push_back(false);

            auto res = abstraction::AlgorithmHelper::eval("convert::DotConverter",
                                                          {},
                                                          params,
                                                          moves,
                                                          abstraction::AlgorithmCategories::AlgorithmCategory::DEFAULT);

            auto& value = dynamic_cast<abstraction::ValueProvider<std::string>&>(*res);

            return QString::fromStdString(value.getValue(true));
        }
        catch (const exception::CommonException& e) {
            std::cerr << e << std::endl;
            return std::nullopt;
        }
        catch (const std::exception& e) {
            return std::nullopt;
        }
    }


    std::optional<QString> toXML(const std::shared_ptr<abstraction::OperationAbstraction>& data, bool indent) {
        try {
            auto res = abstraction::XmlRegistry::getXmlComposerAbstraction(data->getReturnType());
            res->attachInput(data, 0, true, true);
            res->eval();

            auto& value = dynamic_cast<abstraction::ValueProvider<ext::deque<sax::Token>>&>(*res);

            std::string result;
            sax::SaxComposeInterface::composeMemory(result, value.getValue(true));

            auto text = QString::fromStdString(result);

            if (indent) {
                QDomDocument doc;
                if (doc.setContent(text, false))
                    return doc.toString(4);
            }
            return text;

        }
        catch (const exception::CommonException& e) {
            return std::nullopt;
        }
        catch (const std::exception& e) {
            return std::nullopt;
        }
    }

    std::optional<QImage> toPNG(const std::shared_ptr<abstraction::OperationAbstraction>& data) {
        if (auto dot = Converter::toDOT(data)) {
            return GraphvizIntegrator::createImage(*dot, GraphvizIntegrator::PNG);
        }
        else {
            return std::nullopt;
        }
    }

    std::shared_ptr<abstraction::OperationAbstraction> parseXML(const QString& xml) {
        try {
            ext::deque<sax::Token> tokens;
            sax::SaxParseInterface::parseMemory(xml.toStdString(), tokens);
            std::string type = tokens[0].getData();
            std::shared_ptr<abstraction::OperationAbstraction> automaton = abstraction::XmlRegistry::getXmlParserAbstraction(
                    type);

            auto tokensAbstraction = std::make_shared<abstraction::ImmediateValueAbstraction<ext::deque<sax::Token>>>(
                    tokens);
            automaton->attachInput(tokensAbstraction, 0, true, true);
            return automaton;
        }
        catch (const exception::CommonException& e) {
            return nullptr;
        }
        catch (const std::exception& e) {
            return nullptr;
        }
    }

    std::shared_ptr<abstraction::OperationAbstraction> parseText(const QString& txt) {
        ext::vector<std::shared_ptr<abstraction::OperationAbstraction>> params;
        params.push_back(std::make_shared<abstraction::ImmediateValueAbstraction<std::string>>(txt.toStdString()));

        std::vector<bool> moves;
        moves.push_back(true);
        for (const auto& type: { "automaton::Automaton", "grammar::Grammar", "regexp::RegExp"}) {
            try {
                return abstraction::AlgorithmHelper::eval("string::Parse",
                                                          {type},
                                                          params,
                                                          moves,
                                                          abstraction::AlgorithmCategories::AlgorithmCategory::DEFAULT);
            }
            catch (const exception::CommonException& e) {}
            catch (const std::exception& e) {}
        }
        return nullptr;
    }

    std::shared_ptr<abstraction::OperationAbstraction> tryParse(const QString& input) {
        if (auto automaton = Converter::parseXML(input)) {
            return automaton;
        }
        if (auto automaton = Converter::parseText(input)) {
            return automaton;
        }
        return nullptr;
    }

    void saveToImage(const std::shared_ptr<abstraction::OperationAbstraction>& data, const QString& filename) {
        if (auto dot = Converter::toDOT(data)) {
            if (!GraphvizIntegrator::createImageFile(*dot, filename, GraphvizIntegrator::formatFromFilename(filename))) {
                throw std::runtime_error { "Failed to write to file." };
            }
        }
        else {
            throw std::runtime_error { "Failed to convert data to DOT." };
        }
    }
}
