#include "TimeoutAqlTest.hpp"
#include <cstring>
#include <exception/CommonException.h>
#include <parser/Parser.h>

#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>

#define PIPE_RD 0
#define PIPE_WR 1

/* Communication between signal handler and the rest of the program */
int g_Wakeup [ 2 ];  // pipe for wakeup
int g_RecvSignal; // signalled flag

int waitSignalTimeout ( int timeout ) {
	struct timeval tv;
	fd_set rd;

	tv.tv_sec = 0;
	tv.tv_usec = timeout;
	FD_ZERO ( &rd );
	FD_SET ( g_Wakeup [ PIPE_RD ], &rd );

	select ( g_Wakeup [ PIPE_RD ] + 1, &rd, NULL, NULL, &tv );
	return g_RecvSignal;
}

void newSigChild ( int ) {
	char dummy = 0;
	g_RecvSignal = 1;

	// write into the pipe so select can read something, this effectively means that SIGCHILD was raised
	if ( write ( g_Wakeup [ PIPE_WR ], &dummy, 1 ) != 1 )
		throw exception::CommonException ( "TimeoutAqlTest: write() failure (wakeup signalling)" );
}

int aqlTest ( int fd, const ext::vector < std::string > & queries ) {
	cli::Environment environment;

	try {
		for ( const std::string & q : queries ) {
			cli::Parser parser = cli::Parser ( cli::Lexer ( q ) );
			parser.parse ( ) -> run ( environment );
		}
	} catch ( const exception::CommonException & e ) {
		const char * what = e.what ( );
		ssize_t what_len = ( ssize_t ) strlen ( what );
		if ( write ( fd, what, what_len ) != what_len )
			throw exception::CommonException ( "TimeoutAqlTest: write() failure (child to parent communication)" );
		return -1;
	}

	return environment.getResult ( ); /* 0 = OK */
}

std::string readChildOutput ( int fd ) {
	static const size_t BUFSIZE = 64;

	std::string res;
	int rd;
	char buf [ BUFSIZE ];

	while ( ( rd = read ( fd, &buf, BUFSIZE - 1 ) ) > 0 ) {
		res.append ( buf, rd );
	}

	return res;
}

void _TimeoutAqlTest ( const std::chrono::microseconds & timeout, const ext::vector < std::string > & queries ) {
	/* Register SIGCHLD handler */
	struct sigaction act;
	memset ( &act, 0, sizeof ( act ) );
	act . sa_handler = newSigChild;
	sigaction ( SIGCHLD, &act, NULL );

	/* parent-child communication ( for exceptions ) */
	int pipefd [ 2 ];
	if ( pipe ( pipefd ) != 0 )
		throw exception::CommonException ( "TimeoutAqlTest: Failed to initialize pipe (child to parent communication)" );

	/* SIGCHLD was not yet raised, initialize communication pipe */
	g_RecvSignal = 0;
	if ( pipe ( g_Wakeup ) )
		throw exception::CommonException ( "TimeoutAqlTest: Failed to initialize pipe (wakeup signalling)" );

	/* do the forking */
	pid_t x = fork ();
	if ( x < 0 ) {
		FAIL ( "Fork error" );
	} else if ( x == 0 ) {
		/* child, run the test here */
		close ( g_Wakeup [ PIPE_RD ] );
		close ( g_Wakeup [ PIPE_WR ] );
		close ( pipefd [ PIPE_RD ] );

		/* just in case ... */
		close ( 1 );
		close ( 2 );

		exit ( aqlTest ( pipefd [ PIPE_WR ], queries ) );
	}

	close ( pipefd [ PIPE_WR ] );

	/* lets wait the specified time of microseconds, maybe the child will terminate on its own */
	if ( ! waitSignalTimeout ( timeout.count ( ) ) ) {
		/* ... and in case it did not ... */
		kill ( x, SIGTERM );
		while ( ! waitSignalTimeout ( 250000 ) ) /* 1/4 second */
			kill ( x, SIGKILL );
	}

	/* child termination confirmed */
	int status;
	waitpid ( x, &status, 0 );
	close ( g_Wakeup [ PIPE_RD ] );
	close ( g_Wakeup [ PIPE_WR ] );

	std::string childOutput = readChildOutput ( pipefd [ PIPE_RD ] );

	close ( pipefd [ PIPE_RD ] );
	close ( pipefd [ PIPE_WR ] );

	if ( WIFEXITED ( status ) ) {
		// std::cerr << "Child process finished, code " << WEXITSTATUS ( status ) << std::endl;
		INFO ( "AqlTest failure. Trying to execute: " << queries );
		INFO ( "Child output was: >" << childOutput << "<" );
		REQUIRE ( WEXITSTATUS ( status ) == 0 );
	} else if ( WIFSIGNALED ( status ) ) {
		// std::cerr << "Child process signaled, signal ", WTERMSIG ( status ) << std::endl;
		WARN ( "Timeout (" << timeout.count ( ) << " us) reached in test (" << queries << ")" );
	}
}
